//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsResSurfRevLib.h"

//----------------------------------------------------------------
// TODO - _CULLING:
// 1) Set the viewpoint uniform for culling. In which space is the computation, why?

// TODO - _EXPLODE:
// 1) Nothing to do :-)

// TODO - _TESSELLATE:
// 1) Enable backface culling in function "init".
// 2) Notice the difference between data.vert and pass.vert.

// TODO - _NORMALS:
// 1) Another program without geometry shader (just draws the triangles) is setup in function "setupShaders".
// 2) Use that program to draw the geometry and then the normals.
//----------------------------------------------------------------

#define _CULLING		0
#define _EXPLODE		1
#define _TESSELLATE		2
#define _NORMALS		3

#define _GEOM _NORMALS
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader, shaderColor;
VSResSurfRevLib mySurfRev;

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.3f;

// Geometry offset
float offsetX = 0.f, offsetY = 0.f;

float time = 0.f;


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);
	
	// Set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);
  
	vsml->translate(offsetX, offsetY, 0.f);

#if (_GEOM == _NORMALS)
	
#endif

	// Use our shader
	glUseProgram(shader.getProgramIndex());

#if (_GEOM == _CULLING)
	// Set the culling viewpoint in the projection space
#endif

	shader.setUniform("time", time);
	
	mySurfRev.render();

	time += 0.01f;

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;

		case 'm': glEnable(GL_MULTISAMPLE);
			break;

		case 'n': glDisable(GL_MULTISAMPLE);
			 break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for models
	shader.init();

#if (_GEOM == _CULLING)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/data.vert");
	shader.loadShader(VSShaderLib::GEOMETRY_SHADER, "shaders/culling.geom");
#elif (_GEOM == _EXPLODE)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/data.vert");
	shader.loadShader(VSShaderLib::GEOMETRY_SHADER, "shaders/explode.geom");
#elif (_GEOM == _TESSELLATE)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	shader.loadShader(VSShaderLib::GEOMETRY_SHADER, "shaders/tessellate.geom");
#elif (_GEOM == _NORMALS)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	shader.loadShader(VSShaderLib::GEOMETRY_SHADER, "shaders/normals.geom");
#endif

	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/color.frag");

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");
	shader.prepareProgram();

	printf("InfoLog for Model Shader\n%s\n\n", shader.getAllInfoLogs().c_str());

#if (_GEOM == _NORMALS)
	shaderColor.init();

	shaderColor.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/data.vert");
	shaderColor.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/color.frag");

	// set semantics for the shader variables
	shaderColor.setProgramOutput(0, "outputF");
	shaderColor.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shaderColor.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");
	shaderColor.prepareProgram();

	printf("InfoLog for Model Shader\n%s\n\n", shaderColor.getAllInfoLogs().c_str());
#endif
	
	return(shader.isProgramValid());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	//mySurfRev.createCone(1.f, 1.f, 50);
	//mySurfRev.createCylinder(1.f, 1.f, 50);
	mySurfRev.createTorus(0.5f, 1.0f, 50, 50);
	//mySurfRev.createPawn();
	//mySurfRev.createSphere(1.f, 50);

	// Some GL settings
	glEnable(GL_DEPTH_TEST);
	//glEnable(GL_CULL_FACE);
	glEnable(GL_MULTISAMPLE);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::NORMAL, "m_normal");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Geometry shaders");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}

